'''
Aggregate subgradient method for nonsmooth DC optimization.

Created on 14 Jan 2020

@author: Napsu Karmitsa, 

Reference:
A. Bagirov, S. Taheri, K. Joki, N. Karmitsa, M.M. Mäkelä, 
"Aggregate subgradient method for nonsmooth DC optimization", Optimization Letters, 
to appear, 2020.

Acknowledgement:
The work was financially supported by Academy of Finland (Project No. 289500). 

'''
# Load library
import numpy as np

# Define aggsub
def aggsub(x,eps,mit,tau,sig1,delta,sig2,c1,c2):
    """ Aggregate subgradient method for nonsmooth DC optimization.
    
    Input:
        x     - a starting point;
        eps   - optimality tolerance;
        mit   - maximum number of inner iterations;   
        tau   - proximity parameter, tau > 0;
        sig1  - decrease parameter for tau, sig1 in (0,1);
        delta - inner iteration tolerance, delta > 0; 
        sig2  - decrease parameter for delta, sig2 in (0,1];
        c1,c2 - line search parameters, 0 < c2 <= c1 < 1. 
        
    Output:
        x     - a solution obtained;
        f     - the objective value at x;   
        nit   - number of iterations;
        nf    - number of function evaluations;        
        ng    - number of subgradient evaluations;
    """

    # Import DC functions and their subgradients
    import functions
        
    fdiff = -99.0   
    # Initialization
    f1 = functions.f1(x)
    f2 = functions.f2(x)
    fold = f1-f2   # Original function value        
    #print "f1 is", f1
    #print "f2 is", f2
    print "f original is", fold
    print "Computing..."
    nf = 1                    # Number of function evaluations.
    ng = 0                    # Number of subgradient evaluations.         
    nit = 0                   # Number of iterations.
    maxii = max(min(x.size,500),50)    # Maximum number of inner iterations.
    stopii = -1               # reason to stop the inner iteration:
    # stopii =-1 - not stopped yet;
    # stopii = 0 - small gradient: canditate solution;
    # stopii = 1 - decent direction found;
    # stopii = 2 - too many inner iterations without progress.
    small = 1e-5
    nrmnew = 1e+10
         
    # Outer iteration
    while True:
        # Step 1
        nii = 0               # Number of inner iterations.
        dd = np.ones_like(x)/np.sqrt(x.size)  # Initialization of the direction dd.
        g1 = functions.df1(x+tau*dd) # Subgradient of the DC component f1.
        g2 = functions.df2(x) # Subgradient of the DC component f2.
        # No need to recompute g2 if returning from Step 6.
        ng += 1
        sg = g1-g2            # Approx. subgradient of the function.
        asg = sg              # Aggregate subgradient.
        nrmasg = 0.0          # Norm of the aggregate subgradient.
    
        # Inner iteration
        while True:
            nii += 1
        
            # Step 2
            sgdiff=np.dot(sg-asg,sg-asg)
            if sgdiff > small:
                lam = (nrmasg*nrmasg - np.dot(sg,asg))/sgdiff
            else:
                lam = 0.0
                
            if lam > 1.0 or lam < 0.0:
                print "Projecting lambda = ", lam, "to [0,1]" 
                if lam < 0:
                    lam = 0.0
                else:
                    lam = 1.0    
                
                    
            # Step 3
            asg = lam*sg + (1.0-lam)*asg  # The new aggregate subgradient.
            nrmasg = np.linalg.norm(asg)  # Norm of the aggregate subgradient.
            
            if nrmasg < delta:  # Inner iteration termination
                print "nrmasg < delta"
                stopii = 0
                break           # With this we should go to Step 6.
            
            if nii%5==0:
                nrmold = nrmnew
                nrmnew = nrmasg
                if np.abs(nrmold-nrmnew)<0.0001:
                    print "Norm is not changing."
                    stopii = 0
                    break       # With this we should go to Step 6.
            
            # Step 4: Search direction
            dd = - asg / nrmasg
            
            # Step 5
            xtau = x + tau*dd
            f1 = functions.f1(xtau)
            f2 = functions.f2(xtau)
            fnew = f1-f2
            nf += 1
            dt = fnew - fold
            
            if dt > -c1*tau*nrmasg: # Not a descent direction
                g1 = functions.df1(xtau)
                sg = g1-g2
                ng += 1
            else:
                #print "Decent direction found."
                stopii = 1
                break # with this we should go to Step 7.
            
                
            # Additional termination from inner iteration
            if nii > maxii:
                if tau > eps:
                    #print "Too many inner iterations. Adjusting tau."
                    stopii = 2
                else:
                    print "Too many inner iterations. No descent direction found."
                    stopii = 3
                break
        
        # Step 6: Stopping criterion 
        if stopii == 0: # Small norm 
            if tau < eps:
                print "Critical point found"
                break # Critical point found   
            else:
                tau *= sig1
                delta *= sig2
                stopii = -1
                nit += 1
                continue
            
        # Step 7: Line search
        elif stopii == 1: # Descent direction
            step = tau    # Here fnew = f(xtau), fold = f(x)
            count = 0
            while count < 20: # Maximum of 20 search are made
                count +=1
                step *=2.0
                xnew = x + step*dd
                f1 = functions.f1(xnew)
                f2 = functions.f2(xnew)
                nf += 1
                fdiff = f1-f2 - fold
            
                if fdiff <= -c2 * step * nrmasg:
                    xtau = xnew
                    fnew = f1-f2
                else:
                    break
            
            # Step 8: Update step
        
            x = xtau 
            fold = fnew
        
        #print "fdiff, dt",fdiff,dt
        if tau > eps:
            tau *= sig1 # tau too small is not good, needs safeguard 
        elif np.abs(fdiff) < 1e-10:
            print "Termination with small change in function values.",fdiff
            break
            
        delta *= sig2
            
        nit += 1
        # Termination with maximum number of iterations.
        if nit >= mit: # Number of iterations > mit
            print "Termination with maximum number of iterations."
            break
        
        if stopii == 3:
            #print "No decent direction found in inner iteration."
            break
        
        stopii = -1 # To next inner iteration.
            
    f=fold 
    print "norm =", nrmasg
    # Outer iteration ends
    
    return x,f,nit,nf,ng